@@ -374,6 +374,7 @@ sub _connect_unlocked
 		$socket->setsockopt(IPPROTO_TCP, TCP_NODELAY, 1);
 
 		$| = 1; # enable autoflush
+
 		if(defined(&{"MSG_NOSIGNAL"}))
 		{
 			$socket->send('', MSG_NOSIGNAL);
@@ -596,6 +597,7 @@ sub _disconnect_unlocked
 	# destroy receive thread (2/2)
 	if(defined($self->{receive_thread}))
 	{
+		# FIXME: fails with while auto-reconneting "Cannot join self at Tinkerforge/IPConnection.pm line 599, <STDIN> line 1."
 		$self->{receive_thread}->join();
 		$self->{receive_thread} = undef;
 	}
@@ -624,6 +626,7 @@ sub _destroy_socket
 		{
 			$socket->shutdown(2);
 		};
+
 		eval
 		{
 			$socket->close();
@@ -866,16 +869,16 @@ sub enumerate
 
 =item register_callback()
 
-Registers a callback with ID $id to the function named $callback.
+Registers the given $function name with the given $callback_id.
 
 =back
 =cut
 
 sub register_callback
 {
-	my ($self, $function_id, $function_name) = @_;
+	my ($self, $callback_id, $function) = @_;
 
-	$self->{registered_callbacks}->{$function_id} = '\&'.caller.'::'.$function_name;
+	$self->{registered_callbacks}->{$callback_id} = '\&'.caller.'::'.$function;
 }
 
 sub _create_packet_header
@@ -977,7 +980,8 @@ sub _handle_disconnect_by_peer
 
 	$self->{auto_reconnect_allowed} = 1;
 
-	if($disconnect_immediately) {
+	if($disconnect_immediately)
+	{
 		$self->_disconnect_unlocked();
 	}
 
@@ -997,6 +1001,7 @@ sub _ipcon_send
 	}
 
 	my $rc;
+
 	eval
 	{
 		lock(${$self->{send_lock_ref}});
@@ -1011,6 +1016,7 @@ sub _ipcon_send
 			$rc = $self->_get_local_socket()->send($packet);
 		}
 	};
+
 	if(!defined($rc))
 	{
 		$self->_handle_disconnect_by_peer(&DISCONNECT_REASON_ERROR, 0, 1);
@@ -1105,6 +1111,7 @@ sub _get_err_from_data
 	{
 		@bit_arr = split('', unpack('(b8)<', $data_arr[7]));
 		$err = $bit_arr[7].$bit_arr[6];
+
 		return oct("0b$err");
 	}
 
@@ -1126,6 +1133,7 @@ sub _get_payload_from_data
 		{
 			$payload .= $data_arr[$i];
 		}
+
 		return $payload;
 	}
 
@@ -1147,6 +1155,7 @@ sub _handle_packet
 		{
 			$self->{callback_queue}->enqueue([&_QUEUE_PACKET, $packet, undef, undef]);
 		}
+
 		return 1;
 	}
 
@@ -1186,8 +1195,10 @@ sub _handle_packet
 					return 1;
 				}
 			}
+
 			$self->{callback_queue}->enqueue([&_QUEUE_PACKET, $packet, undef, undef]);
 		}
+
 		return 1;
 	}
 
@@ -1412,7 +1423,8 @@ sub _dispatch_meta
 			{
 				$retry = 0;
 
-				if(1) {
+				if(1)
+				{
 					lock(${$self->{socket_lock_ref}});
 
 					if($self->{auto_reconnect_allowed} && !defined($self->{socket_fileno}))
@@ -1421,6 +1433,7 @@ sub _dispatch_meta
 						{
 							$self->_connect_unlocked(1);
 						};
+
 						if($!)
 						{
 							$retry = 1;
@@ -1455,14 +1468,17 @@ sub _dispatch_response
 	my @form_unpack_arr = split(' ', $form_unpack);
 	my @form_unpack_arr_patched = @form_unpack_arr;
 
-	foreach(@form_unpack_arr_patched) {
+	foreach(@form_unpack_arr_patched)
+	{
 		my @form_single_arr = split('', $_);
 
-		if($form_single_arr[0] eq '?' && scalar(@form_single_arr) == 1) {
+		if($form_single_arr[0] eq '?' && scalar(@form_single_arr) == 1)
+		{
 			$form_unpack_arr[$iter] = 'C';
 			$form_unpack_arr_patched[$iter] = 'C';
 		}
-		elsif($form_single_arr[0] eq '?' && scalar(@form_single_arr) > 1) {
+		elsif($form_single_arr[0] eq '?' && scalar(@form_single_arr) > 1)
+		{
 			my $count = $_;
 			$count =~ s/[^\d]//g;
 			$form_unpack_arr_patched[$iter] = 'C' . ceil($count / 8);
@@ -1491,18 +1507,21 @@ sub _dispatch_response
 			my $count = $_;
 			my @form_single_arr = split('', $_);
 
-			if($form_single_arr[0] eq '?') {
+			if($form_single_arr[0] eq '?')
+			{
 				if(scalar(@form_single_arr) > 1)
 				{
 					$count =~ s/[^\d]//g;
 					$return_arr[$iter] = [];
 					my @dummy_bool_array_bits = ();
 
-					for(my $i = 0; $i < ceil($count / 8); $i++) {
+					for(my $i = 0; $i < ceil($count / 8); $i++)
+					{
 						$dummy_bool_array_bits[$i] = $arguments_arr[$copy_from_index + $i];
 					}
 
-					for(my $i = 0; $i < $count; $i++) {
+					for(my $i = 0; $i < $count; $i++)
+					{
 						push(@{$return_arr[$iter]}, (($dummy_bool_array_bits[floor($i / 8)] & (1 << ($i % 8))) != 0));
 					}
 
@@ -1556,7 +1575,8 @@ sub _dispatch_response
 			{
 				if(scalar(@copy_info_arr) > 0)
 				{
-					if($copy_info_arr[0][0] == -1) {
+					if($copy_info_arr[0][0] == -1)
+					{
 						$anon_arr_index++;
 						$arguments_arr_index += $copy_info_arr[0][1];
 						shift(@copy_info_arr);
@@ -1617,14 +1637,7 @@ sub _dispatch_response
 
 			if(defined($fid))
 			{
-				if(defined($uid))
-				{
-					eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}(\@return_arr)");
-				}
-				else
-				{
-					eval("$self->{registered_callbacks}->{$fid}(\@return_arr)");
-				}
+				$self->_dispatch_callback($uid, $fid, \@return_arr);
 			}
 
 			return @return_arr;
@@ -1654,14 +1667,7 @@ sub _dispatch_response
 
 			if(defined($fid))
 			{
-				if(defined($uid))
-				{
-					eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}(\@return_arr)");
-				}
-				else
-				{
-					eval("$self->{registered_callbacks}->{$fid}(\@return_arr)");
-				}
+				$self->_dispatch_callback($uid, $fid, \@return_arr);
 			}
 
 			return @return_arr;
@@ -1689,14 +1695,7 @@ sub _dispatch_response
 
 				if(defined($fid))
 				{
-					if(defined($uid))
-					{
-						eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}(\@return_arr)");
-					}
-					else
-					{
-						eval("$self->{registered_callbacks}->{$fid}(\@return_arr)");
-					}
+					$self->_dispatch_callback($uid, $fid, \@return_arr);
 				}
 
 				return \@return_arr;
@@ -1707,14 +1706,7 @@ sub _dispatch_response
 
 				if(defined($fid))
 				{
-					if(defined($uid))
-					{
-						eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}(\@return_arr)");
-					}
-					else
-					{
-						eval("$self->{registered_callbacks}->{$fid}(\@return_arr)");
-					}
+					$self->_dispatch_callback($uid, $fid, \@return_arr);
 				}
 
 				return \@return_arr;
@@ -1725,14 +1717,7 @@ sub _dispatch_response
 
 				if(defined($fid))
 				{
-					if(defined($uid))
-					{
-						eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}($return_arr[0])");
-					}
-					else
-					{
-						eval("$self->{registered_callbacks}->{$fid}($return_arr[0])");
-					}
+					$self->_dispatch_callback($uid, $fid, $return_arr[0]);
 				}
 
 				return $return_arr[0];
@@ -1743,14 +1728,7 @@ sub _dispatch_response
 
 				if(defined($fid))
 				{
-					if(defined($uid))
-					{
-						eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}(\@return_arr)");
-					}
-					else
-					{
-						eval("$self->{registered_callbacks}->{$fid}(\@return_arr)");
-					}
+					$self->_dispatch_callback($uid, $fid, \@return_arr);
 				}
 
 				return \@return_arr;
@@ -1762,14 +1740,7 @@ sub _dispatch_response
 
 			if(defined($fid))
 			{
-				if(defined($uid))
-				{
-					eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}($unpack_tmp)");
-				}
-				else
-				{
-					eval("$self->{registered_callbacks}->{$fid}($unpack_tmp)");
-				}
+				$self->_dispatch_callback($uid, $fid, $unpack_tmp);
 			}
 
 			return $unpack_tmp;
@@ -1779,14 +1750,7 @@ sub _dispatch_response
 	{
 		if(defined($fid))
 		{
-			if(defined($uid))
-			{
-				eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}()");
-			}
-			else
-			{
-				eval("$self->{registered_callbacks}->{$fid}()");
-			}
+			$self->_dispatch_callback($uid, $fid, undef);
 		}
 
 		return 1;
@@ -1795,6 +1759,161 @@ sub _dispatch_response
 	return 1;
 }
 
+sub _dispatch_callback
+{
+	my ($self, $uid, $fid, $args, $is_array) = @_;
+
+	if(defined($uid))
+	{
+		if(defined($args))
+		{
+			if(ref($args) eq "ARRAY")
+			{
+				if(defined($self->{devices}->{$uid}->{high_level_callbacks}->{-$fid}))
+				{
+					my $hlcb = $self->{devices}->{$uid}->{high_level_callbacks}->{-$fid}; # [roles-by-name, roles-by-index, options, data]
+					my $has_data = 0;
+					my $data = undef;
+					my $length = undef;
+					my $chunk_offset = undef;
+
+					if(defined(@{$hlcb}[2]->{fixed_length}))
+					{
+						$length = @{$hlcb}[2]->{fixed_length};
+					}
+					else
+					{
+						$length = @{$args}[@{$hlcb}[0]->{stream_length}];
+					}
+
+					if(!@{$hlcb}[2]->{single_chunk})
+					{
+						$chunk_offset = @{$args}[@{$hlcb}[0]->{stream_chunk_offset}];
+					}
+					else
+					{
+						$chunk_offset = 0;
+					}
+
+					my $chunk_data = @{$args}[@{$hlcb}[0]->{stream_chunk_data}];
+
+					if(!defined(@{$hlcb}[3])) # no stream in-progress
+					{
+						if($chunk_offset == 0) # stream starts
+						{
+							@{$hlcb}[3] = shared_clone($chunk_data);
+
+							if(scalar(@{@{$hlcb}[3]}) >= $length) # stream complete
+							{
+								$has_data = 1;
+								$data = @{$hlcb}[3];
+
+								# FIXME: splice doesn't support shared arrays, manually drop extra items
+								while(scalar(@{$data}) > $length)
+								{
+									pop(@{$data});
+								}
+
+								@{$hlcb}[3] = undef;
+							}
+						}
+						else # ignore tail of current stream, wait for next stream start
+						{
+						}
+					}
+					else # stream in-progress
+					{
+						if($chunk_offset != scalar(@{@{$hlcb}[3]})) # stream out-of-sync
+						{
+							$has_data = 1;
+							$data = undef;
+							@{$hlcb}[3] = undef;
+						}
+						else # stream in-sync
+						{
+							push(@{@{$hlcb}[3]}, @{$chunk_data});
+
+							if(scalar(@{@{$hlcb}[3]}) >= $length) # stream complete
+							{
+								$has_data = 1;
+								$data = @{$hlcb}[3];
+
+								# FIXME: splice doesn't support shared arrays, manually drop extra items
+								while(scalar(@{$data}) > $length)
+								{
+									pop(@{$data});
+								}
+
+								@{$hlcb}[3] = undef;
+							}
+						}
+					}
+
+					if($has_data && defined($self->{devices}->{$uid}->{high_level_callbacks}->{-$fid}))
+					{
+						my @result = ();
+
+						for(my $i = 0; $i < scalar(@{@{$hlcb}[1]}); ++$i)
+						{
+							if(@{@{$hlcb}[1]}[$i] eq 'stream_chunk_data')
+							{
+								push(@result, $data);
+							}
+							elsif(!defined(@{@{$hlcb}[1]}[$i]))
+							{
+								push(@result, @{$args}[$i]);
+							}
+						}
+
+						eval("$self->{devices}->{$uid}->{registered_callbacks}->{-$fid}(\@result)");
+					}
+				}
+
+				if(defined($self->{devices}->{$uid}->{registered_callbacks}->{$fid}))
+				{
+					# FIXME: for some unknown reason directly passing the reference
+					#        to an array doesn't work with eval. it only works if the
+					#        reference is created locally by eval
+					my @array = @{$args};
+					eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}(\@array)");
+				}
+			}
+			else
+			{
+				eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}($args)");
+			}
+		}
+		else
+		{
+			eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}()");
+		}
+	}
+	else
+	{
+		if(defined($args))
+		{
+			if(ref($args) eq "ARRAY")
+			{
+				# FIXME: for some unknown reason directly passing the reference
+				#        to an array doesn't work with eval. it only works if the
+				#        reference is created locally by eval
+				my @array = @{$args};
+				eval("$self->{registered_callbacks}->{$fid}(\@array)");
+			}
+			else
+			{
+				eval("$self->{registered_callbacks}->{$fid}($args)");
+			}
+		}
+		else
+		{
+			eval("$self->{registered_callbacks}->{$fid}()");
+		}
+	}
+
+	return 1;
+}
+
 sub _dispatch_packet
 {
 	my ($self, $packet) = @_;
@@ -1819,7 +1938,8 @@ sub _dispatch_packet
 		return 1;
 	}
 
-	if(defined($self->{devices}->{$uid}->{registered_callbacks}->{$fid}))
+	if(defined($self->{devices}->{$uid}->{registered_callbacks}->{$fid}) ||
+	   defined($self->{devices}->{$uid}->{high_level_callbacks}->{-$fid}))
 	{
 		$self->_dispatch_response($payload, $self->{devices}->{$uid}->{callback_formats}->{$fid}, $uid, $fid);
 	}
@@ -1844,13 +1964,14 @@ sub _disconnect_probe_thread_subroutine
 
 		if($self->{disconnect_probe_flag}) {
 			my $packet = $self->_create_packet_header(undef, 8, &_FUNCTION_DISCONNECT_PROBE);
-
 			my $rc;
+
 			eval
 			{
 				lock(${$self->{send_lock_ref}});
 
 				$| = 1; # enable autoflush
+
 				if(defined(&{"MSG_NOSIGNAL"}))
 				{
 					$rc = $self->_get_local_socket()->send($packet, MSG_NOSIGNAL);
@@ -1860,6 +1981,7 @@ sub _disconnect_probe_thread_subroutine
 					$rc = $self->_get_local_socket()->send($packet);
 				}
 			};
+
 			if(!defined($rc))
 			{
 				$self->_handle_disconnect_by_peer(&DISCONNECT_REASON_ERROR, $self->{socket_id}, 0);
